"""supported file formats by afmformats

Usage
-----
Directives:

List of supported file formats

   .. afmformats_file_formats::

"""
from docutils.statemachine import ViewList
from docutils.parsers.rst import Directive
from sphinx.util.nodes import nested_parse_with_titles
from docutils import nodes

import afmformats


class Base(Directive):
    required_arguments = 0
    optional_arguments = 0

    def generate_rst(self):
        pass

    def run(self):
        rst = self.generate_rst()

        vl = ViewList(rst, "fakefile.rst")
        # Create a node.
        node = nodes.section()
        node.document = self.state.document
        # Parse the rst.
        nested_parse_with_titles(self.state, vl, node)
        return node.children


class Formats(Base):
    def generate_rst(self):
        rst = []

        rst.append("")
        rst.append(".. csv-table::")
        rst.append("    :header: Format introduced by, Description, "
                   + "Extension, Loader")
        rst.append("    :widths: 15, 5, 20, 60")
        rst.append("    :delim: tab")
        rst.append("")

        # generate a new dict with maker-keys
        loader_dict = {}
        for recipe in afmformats.formats.formats_available:
            # group by loader
            k = recipe["loader"].__name__ + recipe["descr"] + recipe["suffix"]
            loader_dict[k] = {
                "ext": recipe["suffix"],
                "mod": ":func:`{}.{}<{}.{}>`".format(
                    recipe["loader"].__module__.split(".", 1)[1],
                    recipe["loader"].__name__,
                    recipe["loader"].__module__,
                    recipe["loader"].__name__),
                "mak": recipe["maker"],
                "des": recipe["descr"]
            }
        for item in sorted(loader_dict.values(), key=lambda x: x["ext"]):
            rst.append("    {}\t {}\t {} \t {}".format(
                item["mak"], item["des"], item["ext"], item["mod"]))

        rst.append("")

        return rst


def setup(app):
    app.add_directive('afmformats_file_formats', Formats)
    return {'version': '0.2'}   # identifies the version of our extension



"""Test Chiaro text export format"""
import pathlib
import zipfile
import tempfile
import numpy as np

import afmformats


def _load_zip_data_path():
    """Convenience function to find zipped test data path"""
    zpath = (pathlib.Path(__file__).resolve().parent / "data" /
             "fmt-chiaro-txt_AEBP1_Indentation_002.zip")
    # unpack
    arc = zipfile.ZipFile(str(zpath))
    # extract all files to a temporary directory
    edest = tempfile.mkdtemp(prefix=zpath.name)
    arc.extractall(edest)
    zip_temp_path = pathlib.Path(edest)
    data_paths = [r for r in zip_temp_path.rglob("*.txt") if r.is_file()]
    data_path = data_paths[0]
    return data_path


data_path = _load_zip_data_path()


def test_chairo_txt_open_check_data():
    data = afmformats.load_data(data_path)[0]
    assert data.metadata["imaging mode"] == "force-distance"
    assert data["force"][2000] == 9.399999999999999e-11


def test_chairo_txt_detect():
    recipe = afmformats.formats.get_recipe(data_path)
    assert recipe.descr == "exported by Optics11 Chiaro Indenter"
    assert recipe.maker == "Optics11 Life"


def test_chairo_txt_data_columns():
    data = afmformats.load_data(data_path)[0]

    assert "force" in data
    assert "height (piezo)" in data
    assert "height (measured)" in data
    assert "time" in data
    assert "segment" in data
    assert "tip position" in data  # already calculated when loaded


def test_chairo_txt_metadata():
    data = afmformats.load_data(data_path)[0]

    metadata = data.metadata
    avail_keys = ["date", "duration", "spring constant", "time",
                  "speed approach", "speed retract", "segment count"]
    for key in avail_keys:
        assert key in metadata


def test_chiaro_txt_segments():
    data = afmformats.load_data(data_path)[0]

    assert data["force"].size == 13827
    # maximum force is at end of 1st segment
    assert np.sum(data["segment"] == 0) == np.argmax(data["force"]) + 1


if __name__ == "__main__":
    # Run all tests
    _loc = locals()
    for _key in list(_loc.keys()):
        if _key.startswith("test_") and hasattr(_loc[_key], "__call__"):
            _loc[_key]()



import numpy as np
import pathlib

__all__ = ["load_txt"]


def parse_metadata(file_metadata):
    """Parse the raw text file metadata information."""
    file_metadata = [m for m in file_metadata if m != '\n']
    file_metadata = [_d.strip('\n').split('\t') for _d in file_metadata]

    # handle each weirdly written piece of metadata
    notes = {}
    for sublist in file_metadata:
        if len(sublist) % 2 == 0:
            even, odd = sublist[::2], sublist[1::2]
            notes.update(dict(zip(even, odd)))
        elif len(sublist) == 1:

            if sublist[0].count(":") == 1:
                notes[sublist[0].split(":")[0]] = sublist[0].split(":")[1]
                continue

            # try to split by whitespace and units within parentheses
            split_sublist = sublist[0].split(' ')
            if len(split_sublist) % 2 == 0:
                notes[split_sublist[0]] = split_sublist[1]
            elif len(split_sublist) == 1:
                # no value
                notes[sublist[0]] = sublist[0]
            else:
                if "(" in sublist[0] and ")" in sublist[0]:
                    assert "(" in split_sublist[1]  # should be the second part
                    notes[split_sublist[0] + " " + split_sublist[1]] = \
                        split_sublist[2]

    # cleanup trailing whitespaces
    for key, val in notes.items():
        notes[key] = val.strip()
    return notes


def open_check_content(path):
    # encoding found with `chardet.detect`
    with path.open(encoding='ISO-8859-1') as fd:
        txtdata = fd.readlines()

    valid = False
    ind = None
    for line in txtdata:
        if "Device:	Chiaro" in line:
            valid = True
        elif "Time (s)\tLoad" in line:
            ind = txtdata.index(line)
    if not valid:
        raise ValueError

    _metadata = txtdata[:ind]
    _columns = txtdata[ind].strip('\n').split('\t')
    _data = txtdata[ind + 1:]
    _data_split = [_d.strip('\n').split('\t') for _d in _data]
    _data = np.asarray(_data_split).astype(float)
    assert len(_columns) == _data.shape[1]

    return _metadata, _columns, _data


def detect_txt(path):
    """File should be plain text"""
    valid = False
    try:
        _metadata, _columns, _data = open_check_content(path)
    except (ValueError, IndexError):
        pass
    else:
        if np.issubdtype(_data.dtype, np.floating):
            valid = True
    return valid


def load_txt(path, callback=None, meta_override=None):
    """Load text files exported by the Optics11 Chiaro Indenter.

    Parameters
    ----------
    path: str or pathlib.Path or io.TextIOBase
        path to an chiaro-exported .txt file
    callback: callable
        function for progress tracking; must accept a float in
        [0, 1] as an argument.
    meta_override: dict
        if specified, contains key-value pairs of metadata that
        are used when loading the files
        (see :data:`afmformats.meta.META_FIELDS`)
    """
    if meta_override is None:
        meta_override = {}

    _metadata, _columns, _data = open_check_content(path)

    data = {
        "time": _data[:, 0],
        "force": _data[:, 1] * 1e-6,  # load (uN)
        "height (measured)": _data[:, 2] * 1e-9,  # Indentation (nm)
        "height (piezo)": _data[:, 4] * 1e-9,  # Piezo (nm)
    }
    data["height (measured)"] *= -1
    data["height (piezo)"] *= -1

    cantilever = _data[:, 3] * 1e-9  # Cantilever (nm)
    cantilever *= -1
    # we use the file tip position, as nanite computes it incorrectly
    data["tip position"] = data["height (piezo)"] - cantilever

    max_force_ind = np.argmax(data["force"])

    data["segment"] = np.zeros(_data.shape[0], dtype=np.uint8)
    data["segment"][:max_force_ind + 1] = 0
    data["segment"][max_force_ind + 1:] = 1

    notes = parse_metadata(_metadata)

    # Metadata
    metadata = {}
    metadata["duration"] = float(np.max(data["time"]))
    # acquisition
    metadata["imaging mode"] = "force-distance"
    metadata["feedback mode"] = "contact"
    metadata["setpoint"] = float(notes["Max load (µN)"])
    metadata["spring constant"] = float(notes["k (N/m)"])
    metadata["z range"] = float(notes["Piezo position (nm) (Measured)"])

    # dataset
    metadata["enum"] = 0
    metadata["speed approach"] = float(notes["Piezo speed  (µm/s)"])
    metadata["speed retract"] = float(notes["Piezo speed  (µm/s)"])

    # setup
    metadata["instrument"] = notes["Device:"]
    metadata["software"] = notes["Device:"]
    metadata["software version"] = notes["Software version"]

    # storage
    metadata["path"] = pathlib.Path(path)
    metadata["date"] = notes["Date"]
    metadata["point count"] = data["time"].shape[0]
    metadata["time"] = notes["Time"]

    metadata.update(meta_override)

    dd = {"data": data,
          "metadata": metadata}

    if callback is not None:
        callback(1)

    return [dd]


recipe_chiaro_txt = {
    "descr": "exported by Optics11 Chiaro Indenter",
    "detect": detect_txt,
    "loader": load_txt,
    "suffix": ".txt",
    "modalities": ["force-distance"],
    "maker": "Optics11 Life",
}



import pathlib

import numpy as np


__all__ = ["load_my_format"]


def load_my_format(path, callback=None, meta_override=None):
    """Loads AFM data from my format

    This is the main function for loading your file format. Please
    add a description here.

    Parameters
    ----------
    path: str or pathlib.Path or io.TextIOBase
        path to a .tab file
    callback: callable
        function for progress tracking; must accept a float in
        [0, 1] as an argument.
    meta_override: dict
        if specified, contains key-value pairs of metadata that
        are used when loading the files
        (see :data:`afmformats.meta.META_FIELDS`)
    """
    if meta_override is None:
        meta_override = {}

    path = pathlib.Path(path)
    # Here you would start parsing your data and metadata from `path`
    # You should specify as many metadata keys as possible. See
    # afmformats.meta.DEF_ALL for a list of valid keys.
    metadata = {"path": path}
    # Valid column names are defined in afmformats.afm_data.known_columns.
    data = {"force": np.linspace(1e-9, 5e-9, 100),
            "height (measured)": np.linspace(2e-6, -1e-6, 100)}

    metadata.update(meta_override)
    dd = {"data": data,
          "metadata": metadata}

    if callback is not None:
        callback(1)

    # You may also return a list with more items in case the file format
    # contains more than one curve.
    return [dd]


recipe_myf = {
    "descr": "A short description",
    "loader": load_my_format,
    "suffix": ".myf",
    "modality": "force-distance",
    "maker": "designer of file format",
}



"""Double-exponential fit to stress-relaxation data

This example reproduces the first entry in figure 5d (cell1) of
:cite:`Battistella_2022`.

Please note that in the original manuscript, not all curves were
used in the final figure. Some curves were excluded based on
curve quality. The data were kindly provided by Alice Battistella.
"""

import pathlib
import shutil
import tempfile

import afmformats

from lmfit.models import ConstantModel, ExponentialModel
import matplotlib.pylab as plt
from matplotlib.ticker import ScalarFormatter
import numpy as np
import pandas
import seaborn as sns
sns.set_theme(style="whitegrid", palette="muted")


# extract the data to a temporary directory
data_path = pathlib.Path(__file__).parent / "data"
wd_path = pathlib.Path(tempfile.mkdtemp())
shutil.unpack_archive(data_path / "10.1002_btm2.10294_fig5d-cell1.zip",
                      wd_path)

# scale conversion constants
xsc = 1
ysc = 1e9
xl = "time [s]"
yl = "force [nN]"

# data extraction
data = []
for path in wd_path.glob("*.jpk-force"):
    data += afmformats.load_data(path)

# this is where the data for the swarm plot is stored
rdat = {
    "tau": [],
    "component": [],
    "cell": [],
}

data_fits = []

for di in data:
    # use intermediate `intr` segment data
    x = di.intr["time"] * xsc
    y = di.intr["force"] * ysc

    # fitting with double-exponential
    const = ConstantModel()
    exp_1 = ExponentialModel(prefix='exp1_')
    exp_2 = ExponentialModel(prefix='exp2_')

    pars = exp_1.guess(y, x=x)
    pars.update(exp_2.guess(y, x=x))
    pars["exp1_decay"].set(value=1, min=0.1)
    pars["exp2_decay"].set(value=8, min=2)
    pars.update(const.guess(y, x=x))
    pars["c"].set(value=np.min(y))

    mod = const + exp_1 + exp_2

    init = mod.eval(pars, x=x)
    out = mod.fit(y, pars, x=x)
    data_fits.append([x, out.best_fit])

    rdat["tau"].append(out.params["exp1_decay"].value)
    rdat["component"].append("tau1")
    rdat["cell"].append("cell1, t=0")

    rdat["tau"].append(out.params["exp2_decay"].value)
    rdat["component"].append("tau2")
    rdat["cell"].append("cell1, t=0")


fig = plt.figure(figsize=(8, 4))

# plot all stress-relaxation curves
ax1 = plt.subplot(121, title="double-exponential fits")
for ii in range(len(data)):
    di = data[ii]
    ax1.plot(di["time"]*xsc, di["force"]*ysc, color="k", lw=3)
    xf, yf = data_fits[ii]
    ax1.plot(xf, yf, color="r", lw=1)
ax1.set_xlabel(xl)
ax1.set_ylabel(yl)

df = pandas.DataFrame(data=rdat)
# Draw a categorical scatterplot to show each observation
ax2 = plt.subplot(122, title="comparison of decay times")
ax2 = sns.swarmplot(data=df, x="cell", y="tau", hue="component")
ax2.set_yscale('log', base=2)
ax2.yaxis.set_major_formatter(ScalarFormatter())
ax2.set_yticks([0.5, 1, 2, 4, 8, 16, 32])
ax2.set(xlabel="", ylabel="tau [s]")

plt.tight_layout()

plt.show()
